@@ -9,21 +9,21 @@
 // or else upgrading old project file will not work
 // reliably.
 // TODO QGIS 4: remove in favor of _QGIS_VERSION
-#define VERSION "3.20.3-Odense"
-#define _QGIS_VERSION "3.20.3-Odense"
+#define VERSION "3.22.1-Białowieża"
+#define _QGIS_VERSION "3.22.1-Białowieża"
 
 //used in vim src/core/qgis.cpp
 //The way below should work but it resolves to a number like 0110 which the compiler treats as octal I think
 //because debuggin it out shows the decimal number 72 which results in incorrect version status.
 //As a short term fix I (Tim) am defining the version in top level cmake. It would be good to
 //reinstate this more generic approach below at some point though
-//#define VERSION_INT 3203
+//#define VERSION_INT 3221
 // TODO QGIS 4: Remove in favor of _QGIS_VERSION_INT
-#define VERSION_INT 32003
-#define _QGIS_VERSION_INT 32003
-#define ABISYM(x) x ## 32003
+#define VERSION_INT 32201
+#define _QGIS_VERSION_INT 32201
+#define ABISYM(x) x ## 32201
 //used in main.cpp and anywhere else where the release name is needed
-#define RELEASE_NAME "Odense"
+#define RELEASE_NAME "Białowieża"
 
 #define QGIS_PLUGIN_SUBDIR "lib/qgis/plugins"
 #define QGIS_DATA_SUBDIR "share/qgis"
@@ -31,7 +31,7 @@
 #define QGIS_LIB_SUBDIR "lib"
 #define QGIS_QML_SUBDIR "qml"
 #define CMAKE_INSTALL_PREFIX "/usr/local"
-#define CMAKE_SOURCE_DIR "/usr/obj/ports/qgis-3.20.3/qgis-3.20.3"
+#define CMAKE_SOURCE_DIR "/usr/obj/ports/qgis-3.22.1/qgis-3.22.1"
 
 #define QGIS_SERVER_MODULE_SUBDIR "lib/qgis/server"
 
@@ -46,10 +46,10 @@
 
 #define QT_PLUGINS_DIR "/usr/local/lib/qt5/plugins"
 
-#define PYTHON_VERSION "3.8.12"
+#define PYTHON_VERSION "3.9.9"
 #define PYTHON_VERSION_MAJOR "3"
-#define PYTHON_VERSION_MINOR "8"
-#define PYTHON_VERSION_PATCH "12"
+#define PYTHON_VERSION_MINOR "9"
+#define PYTHON_VERSION_PATCH "9"
 
 /* #undef USING_NMAKE */
 
@@ -69,6 +69,8 @@
 
 #define HAVE_SERVER_PYTHON_PLUGINS
 
+#define HAVE_OAUTH2_PLUGIN
+
 /* #undef HAVE_OPENCL */
 /* #undef OPENCL_USE_NEW_HEADER */
 
