@@ -26,10 +26,11 @@
 class QgsGeometry;
 class QgsAbstractGeometry;
 class QgsCurve;
+struct geod_geodesic;
 
 /**
  * \ingroup core
- * A general purpose distance and area calculator, capable of performing ellipsoid based calculations.
+ * \brief A general purpose distance and area calculator, capable of performing ellipsoid based calculations.
  *
  * Measurements can either be performed on existing QgsGeometry objects, or using
  * lists of points.
@@ -45,6 +46,8 @@ class QgsCurve;
  * Usually, the measurements returned by QgsDistanceArea are in meters. If no valid
  * ellipsoid is set, then the units may not be meters. The units can be retrieved
  * by calling lengthUnits() and areaUnits().
+ *
+ * Internally, the GeographicLib library is used to calculate all ellipsoid based measurements.
 */
 class CORE_EXPORT QgsDistanceArea
 {
@@ -52,6 +55,11 @@ class CORE_EXPORT QgsDistanceArea
 
     //! Constructor
     QgsDistanceArea();
+    ~QgsDistanceArea();
+
+    //! Copy constructor
+    QgsDistanceArea( const QgsDistanceArea &other );
+    QgsDistanceArea &operator=( const QgsDistanceArea &other );
 
     /**
      * Returns whether calculations will use the ellipsoid. Calculations will only use the
@@ -71,14 +79,23 @@ class CORE_EXPORT QgsDistanceArea
     /**
      * Returns the source spatial reference system.
      * \see setSourceCrs()
+     * \see ellipsoidCrs()
      */
     QgsCoordinateReferenceSystem sourceCrs() const { return mCoordTransform.sourceCrs(); }
 
+    /**
+     * Returns the ellipsoid (destination) spatial reference system.
+     * \see sourceCrs()
+     * \see ellipsoid()
+     * \since QGIS 3.6
+     */
+    QgsCoordinateReferenceSystem ellipsoidCrs() const { return mCoordTransform.destinationCrs(); }
+
     /**
      * Sets the \a ellipsoid by its acronym. Known ellipsoid acronyms can be
      * retrieved using QgsEllipsoidUtils::acronyms().
      * Calculations will only use the ellipsoid if a valid ellipsoid has been set.
-     * \returns true if ellipsoid was successfully set
+     * \returns TRUE if ellipsoid was successfully set
      * \see ellipsoid()
      * \see willUseEllipsoid()
      */
@@ -87,7 +104,7 @@ class CORE_EXPORT QgsDistanceArea
     /**
      * Sets ellipsoid by supplied radii. Calculations will only use the ellipsoid if
      * a valid ellipsoid been set.
-     * \returns true if ellipsoid was successfully set
+     * \returns TRUE if ellipsoid was successfully set
      * \see ellipsoid()
      * \see willUseEllipsoid()
      */
@@ -98,6 +115,7 @@ class CORE_EXPORT QgsDistanceArea
      * ellipsoid if a valid ellipsoid has been set.
      * \see setEllipsoid()
      * \see willUseEllipsoid()
+     * \see ellipsoidCrs()
      */
     QString ellipsoid() const { return mEllipsoid; }
 
@@ -224,7 +242,7 @@ class CORE_EXPORT QgsDistanceArea
      * \param distance distance to format
      * \param decimals number of decimal places to show
      * \param unit unit of distance
-     * \param keepBaseUnit set to false to allow conversion of large distances to more suitable units, e.g., meters to
+     * \param keepBaseUnit set to FALSE to allow conversion of large distances to more suitable units, e.g., meters to
      * kilometers
      * \returns formatted distance string
      * \see formatArea()
@@ -237,7 +255,7 @@ class CORE_EXPORT QgsDistanceArea
      * \param area area to format
      * \param decimals number of decimal places to show
      * \param unit unit of area
-     * \param keepBaseUnit set to false to allow conversion of large areas to more suitable units, e.g., square meters to
+     * \param keepBaseUnit set to FALSE to allow conversion of large areas to more suitable units, e.g., square meters to
      * square kilometers
      * \returns formatted area string
      * \see formatDistance()
@@ -271,37 +289,82 @@ class CORE_EXPORT QgsDistanceArea
 
     /**
      * Given a location, an azimuth and a distance, computes the
-     * location of the projected point. Based on Vincenty's formula
-     * for the geodetic direct problem as described in "Geocentric
-     * Datum of Australia Technical Manual", Chapter 4.
+     * location of the projected point.
+     *
      * \param p1 - location of first geographic (latitude/longitude) point as degrees.
      * \param distance - distance in meters.
      * \param azimuth - azimuth in radians, clockwise from North
      * \return p2 - location of projected point as longitude/latitude.
-     * \note code (and documentation) taken from rttopo project
-     * https://git.osgeo.org/gogs/rttopo/librttopo
-     * - spheroid_project.spheroid_project(...)
-     * -  Valid bounds checking for degrees (latitude=+- 85.05115) is based values used for
-     * -> 'WGS84 Web Mercator (Auxiliary Sphere)' calculations
-     * --> latitudes outside these bounds cause the calculations to become unstable and can return invalid results
+     *
      * \since QGIS 3.0
      */
     QgsPointXY computeSpheroidProject( const QgsPointXY &p1, double distance = 1, double azimuth = M_PI_2 ) const;
 
-  private:
+    /**
+     * Calculates the geodesic line between \a p1 and \a p2, which represents the shortest path on the
+     * ellipsoid between these two points.
+     *
+     * The ellipsoid settings defined on this QgsDistanceArea object will be used during the calculations.
+     *
+     * \a p1 and \a p2 must be in the sourceCrs() of this QgsDistanceArea object. The returned line
+     * will also be in this same CRS.
+     *
+     * The \a interval parameter gives the maximum distance between points on the computed line.
+     * This argument is always specified in meters. A shorter distance results in a denser line,
+     * at the cost of extra computing time.
+     *
+     * If the geodesic line crosses the antimeridian (+/- 180 degrees longitude) and \a breakLine is TRUE, then
+     * the line will be split into two parts, broken at the antimeridian. In this case the function
+     * will return two lines, corresponding to the portions at either side of the antimeridian.
+     *
+     * \since QGIS 3.6
+     */
+    QVector<QVector<QgsPointXY> > geodesicLine( const QgsPointXY &p1, const QgsPointXY &p2, double interval, bool breakLine = false ) const;
 
     /**
-     * Calculates distance from two points on ellipsoid
-     * based on inverse Vincenty's formulae
+     * Calculates the latitude at which the geodesic line joining \a p1 and \a p2 crosses
+     * the antimeridian (longitude +/- 180 degrees).
      *
-     * Points \a p1 and \a p2 are expected to be in degrees and in currently used ellipsoid
+     * The ellipsoid settings defined on this QgsDistanceArea object will be used during the calculations.
      *
-     * \returns distance in meters
-     * \note if course1 is not NULL, bearing (in radians) from first point is calculated
-     * (the same for course2)
+     * \a p1 and \a p2 must be in the ellipsoidCrs() of this QgsDistanceArea object. The returned latitude
+     * will also be in this same CRS.
+     *
+     * \param p1 Starting point, in ellipsoidCrs()
+     * \param p2 Ending point, in ellipsoidCrs()
+     * \param fractionAlongLine will be set to the fraction along the geodesic line joining \a p1 to \a p2 at which the antimeridian crossing occurs.
+     *
+     * \returns the latitude at which the geodesic crosses the antimeridian
+     * \see splitGeometryAtAntimeridian()
+     *
+     * \since QGIS 3.6
      */
-    double computeDistanceBearing( const QgsPointXY &p1, const QgsPointXY &p2,
-                                   double *course1 = nullptr, double *course2 = nullptr ) const;
+    double latitudeGeodesicCrossesAntimeridian( const QgsPointXY &p1, const QgsPointXY &p2, double &fractionAlongLine SIP_OUT ) const;
+
+    /**
+     * Splits a (Multi)LineString \a geometry at the antimeridian (longitude +/- 180 degrees).
+     * The returned geometry will always be a multi-part geometry.
+     *
+     * Whenever line segments in the input geometry cross the antimeridian, they will be
+     * split into two segments, with the latitude of the breakpoint being determined using a geodesic
+     * line connecting the points either side of this segment.
+     *
+     * The ellipsoid settings defined on this QgsDistanceArea object will be used during the calculations.
+     *
+     * \a geometry must be in the sourceCrs() of this QgsDistanceArea object. The returned geometry
+     * will also be in this same CRS.
+     *
+     * If \a geometry contains M or Z values, these will be linearly interpolated for the new vertices
+     * created at the antimeridian.
+     *
+     * \note Non-(Multi)LineString geometries will be returned unchanged.
+     *
+     * \see latitudeGeodesicCrossesAntimeridian()
+     * \since QGIS 3.6
+     */
+    QgsGeometry splitGeometryAtAntimeridian( const QgsGeometry &geometry ) const;
+
+  private:
 
     /**
      * Calculates area of polygon on ellipsoid
@@ -315,7 +378,7 @@ class CORE_EXPORT QgsDistanceArea
      * Precalculates some values
      * (must be called always when changing ellipsoid)
      */
-    void computeAreaInit();
+    void computeAreaInit() const;
 
     void setFromParams( const QgsEllipsoidUtils::EllipsoidParameters &params );
 
@@ -335,24 +398,14 @@ class CORE_EXPORT QgsDistanceArea
     //! ellipsoid parameters
     double mSemiMajor, mSemiMinor, mInvFlattening;
 
-    // utility functions for polygon area measurement
+    mutable std::unique_ptr< geod_geodesic > mGeod;
 
-    double getQ( double x ) const;
-    double getQbar( double x ) const;
+    // utility functions for polygon area measurement
 
     double measure( const QgsAbstractGeometry *geomV2, MeasureType type = Default ) const;
     double measureLine( const QgsCurve *curve ) const;
     double measurePolygon( const QgsCurve *curve ) const;
 
-    // temporary area measurement stuff
-
-    double m_QA, m_QB, m_QC;
-    double m_QbarA, m_QbarB, m_QbarC, m_QbarD;
-    double m_AE;  /* a^2(1-e^2) */
-    double m_Qp;  /* Q at the north pole */
-    double m_E;   /* area of the earth */
-    double m_TwoPI;
-
 };
 
 #endif
